"""Test the reel utility functions."""
import pathlib  # trio.Path.home() gives error

from trio import Path

import reel.tools


def test_import():
    """Import the utility functions."""
    assert reel.tools


async def test_resolve():
    """Resolve file paths to canonical form as a str."""
    assert await reel.tools.resolve('.') == str(await Path('.').resolve())
    assert await reel.tools.resolve('~') == str(pathlib.Path.home())

    parent = await reel.tools.resolve('../')
    assert parent == str(await Path('..').resolve())
    assert parent == str((await Path('.').resolve()).parent)



"""This is reel.output."""
import trio

__all__ = ['Input', 'Output', 'Through']


class Through:
    """."""

    def __init__(self):
        """."""

    def yup(self):
        """."""

    def bup(self):
        """."""


class Input(trio.abc.SendStream):
    """."""

    def __init__(self, send_stream):
        """Store the real stream."""
        self._send_stream = send_stream

    async def __aenter__(self):
        """Enter the stream context."""
        await self._send_stream.__aenter__()
        return self

    async def __aexit__(self, *args):
        """Exit the stream context."""
        await self._send_stream.__aexit__(*args)

    async def aclose(self):
        """Implement trio.AsyncResource and behave nicely."""
        await self._send_stream.aclose()

    async def send_all(self, data):
        """Impl."""
        await self._send_stream.send_all(data)

    async def wait_send_all_might_not_block(self):
        """Impl."""
        await self._send_stream.wait_send_all_might_not_block()


class Output(trio.abc.ReceiveStream):
    """An output stream."""

    def __init__(self, receive_stream):
        """Store the real stream."""
        self._receive_stream = receive_stream

    async def __aenter__(self):
        """Enter the stream context."""
        await self._receive_stream.__aenter__()
        return self

    async def __aexit__(self, *args):
        """Exit the stream context."""
        await self._receive_stream.__aexit__(*args)

    async def aclose(self):
        """Implement trio.AsyncResource and behave nicely."""
        await self._receive_stream.aclose()

    async def receive_some(self, max_bytes):
        """..."""
        return await self._receive_stream.receive_some(max_bytes)

    async def read_bytes(self, limit=2048):
        """..."""
        _output = b''
        # async with self._receive_stream as out:
        print('before True---------------------------')
        while True:
            print('after True---------------------------')
            chunk = await self._receive_stream.receive_some(limit)
            print(f'after chundk-------f{len(chunk)}--------------------')
            if not chunk:
                break
            _output += chunk
            if limit:
                if len(_output) >= limit:
                    break
        return _output



"""Use the sox play utility."""
from reel.proc import Destination


def speaker():
    """Send audio to the speaker."""
    flags = ['-t', 'raw',
             '-r', '44.1k',
             '-e', 'signed-integer',
             '-b', '16',
             '--endian', 'little',
             '-c', '2',
             '-']
    return Destination('play', xconf=flags).receive()



"""Some commonly used test fixtures."""
import os

# import numpy as np
import pytest

# from reel.proc import Source


def set_env(env):
    """Add some variables to the environment."""
    for key in env:
        os.environ[key] = env[key]


def unset_env(env):
    """Remove some variables from the environment."""
    for key in env:
        del os.environ[key]


@pytest.fixture(scope="session")
def env_home(tmp_path_factory):
    """Create an xdg environment with usable home directories."""
    return {
        'XDG_CONFIG_HOME': str(tmp_path_factory.mktemp('xdg_config')),
        'XDG_CACHE_HOME': str(tmp_path_factory.mktemp('xdg_cache')),
        'XDG_DATA_HOME': str(tmp_path_factory.mktemp('xdg_data')),
        'XDG_RUNTIME_DIR': str(tmp_path_factory.mktemp('xdg_runtime')),
    }


@pytest.fixture(scope="session")
def config_icecast(tmp_path_factory):
    """Return variables to populate an icecast.xml config file."""
    return dict(
        location='Neptune',
        admin_email='sushi@trident.sea',
        password='hack-it-up',
        hostname='127.0.0.1',
        port='8555',
        logdir=str(tmp_path_factory.mktemp('icecast_log')),
    )


@pytest.fixture(scope="function")
def audio_dir(tmp_path_factory):
    """Create a directory with different types of audio files."""
    rootdir = tmp_path_factory.mktemp('audio_dir')
    # output_wav = np.cos(2 * np.pi * 440 * np.arange(0, 5, 1 / 44100))
    # ... save the file ...
    # output_flac = Source('flac output.wav')
    return rootdir



"""The runtime configuration environment.

The config module reads settings from environment variables and
assigns defaults when appropriate.

"""
import os

from trio import Path

from reel.tools import resolve


async def get_package_dir():
    """Return the path to this package, assuming this module is top-level."""
    return (await Path(__file__).resolve()).parent


async def get_package_name():
    """Return the name of this package."""
    return (await get_package_dir()).name


async def get_xdg_home(choice=None):
    """Return the root configuration directories for this environment."""
    home_env = {
        'XDG_CONFIG_HOME': await resolve('~/.config'),
        'XDG_CACHE_HOME': await resolve('~/.cache'),
        'XDG_DATA_HOME': await resolve('~/.local/share'),
        'XDG_RUNTIME_DIR': await resolve('~/.local/run'),
    }

    # Override the defaults if already set in an environment variable.
    for key in home_env:
        home_env[key] = os.environ.get(key, home_env[key])

    if choice:
        return home_env[choice]
    return home_env


async def get_xdg_config_dir(app=None, feature=None):
    """Return a config directory for this app.

    Create the directory if it does not exist.

    """
    if app is None:
        app = await get_package_name()
    config_home = Path(await get_xdg_home('XDG_CONFIG_HOME'))
    config_dir = config_home / app
    if not await config_dir.exists():
        await config_dir.mkdir()
    if feature:
        config_dir = config_dir / feature
    if not await config_dir.exists():
        await config_dir.mkdir()
    return config_dir


async def get_config(config_path, template_file, **xconf):
    """Create a config file from a template."""
    config_file = config_path / template_file
    if not await config_file.exists():
        default_file = (await get_package_dir()) / 'templates' / template_file
        formatted = (await default_file.read_text()).format(**xconf)
        await config_file.write_text(formatted)
    return config_file



# pylint: disable=W0611, W0621
"""Test the reel subprocess control module."""
import os
from pathlib import Path

import trio

import reel
from reel.proc import LIMIT, Source
from tests.fixtures import audio_dir


async def test_version():
    """Run a process and read stdout."""
    src = Source('python   -m   reel.cli  -v')
    assert isinstance(src._command, list)  # pylint: disable=protected-access
    assert src.cmd == 'python -m reel.cli -v'
    assert src._env == os.environ  # pylint: disable=protected-access
    assert src.err == ''
    assert src._proc is None  # pylint: disable=protected-access
    assert src.status is None
    assert src._limit == LIMIT  # pylint: disable=protected-access
    assert src._nurse is None  # pylint: disable=protected-access

    version = await src.read_text()
    assert src.status == 0
    assert src.err == ''
    assert isinstance(src._proc,  # pylint: disable=protected-access
                      trio.subprocess.Process)
    assert version == reel.__version__


async def test_pwd():
    """Test a familiar shell command."""
    assert str(Path('.').resolve()) == await Source('pwd').read_text()


async def test_stderr():
    """Run a process and read stderr."""
    not_here = Source('ls /not_here_i_hope')
    assert not await not_here.read_text()  # no output
    assert not_here.status  # error present
    assert not_here.err  # some error about a missing file
    assert 'not_here' in not_here.err


async def test_limit_output():
    """Make sure streaming output from stdout & stderr is limited."""
    infinity = await Source('cat /dev/zero').read_text()
    assert len(infinity) >= LIMIT
    assert len(infinity) <= 2 * LIMIT


async def test_stdin():
    """Send some input to a command."""
    china = await Source('cat').read_text('sunflower')
    assert china == 'sunflower'


# async def test_source_stream():
#     """Process a big file of raw audio."""
#     cmd = '''ffmpeg -ac 2 -i xmas.flac
#              -f s16le -ar 44.1k -acodec pcm_s16le -'''
#     async with Source(cmd) as xmas:
#         async with await xmas.stream() as src:
#             assert isinstance(src, trio.abc.ReceiveStream)
#             size = 0
#             while True:
#                 chunk = await src.receive_some(16384)
#                 if not chunk:
#                     break
#                 size += len(chunk)
#             assert size == 57684912


async def test_short_context():
    """Run a short command with i / o in context managers."""
    async with await Source('cat').stream('china') as sun:
        assert isinstance(sun, reel.io.Output)
        assert len(await sun.read_bytes()) == 5


async def test_audio_dir(audio_dir):
    """Get test audio files."""
    assert audio_dir.exists()
    # assert (audio_dir / 'output.wav').exists()


# async def test_log_stderr():
#     """Read the messages sent from stderr."""
#     with trio.move_on_after(5):
#         cmd = '''ffmpeg -ac 2 -i http://ice1.somafm.com/groovesalad-256-mp3
#                  -f s16le -ar 44.1k -acodec pcm_s16le -'''
#         async with Source(cmd) as soma:
#             soma.log_err(



"""Command line interface for tapedeck."""
import logging
import os

import trio
import trio_click as click

import reel

# Override the log level with an environment variable.
LOGGING_LEVEL = os.environ.get('PL_LOGGING_LEVEL', 'ERROR')
logging.basicConfig(level=LOGGING_LEVEL)


@click.command()
@click.option('-v', '--version', is_flag=True,
              help='Print the version number and exit.')
async def main(**kwargs: str) -> None:
    """Run reel from the command line."""
    await trio.sleep(0.1)
    if kwargs['version']:
        click.echo(reel.__version__)

if __name__ == '__main__':
    trio.run(main())



# pylint: disable=W0401, W0611, W0614, W0621
"""Test the API."""
import reel
from reel.config import (
    get_config,
    get_xdg_config_dir,
)
from reel.io import *
from reel.proc import *
from reel.tools import *
from tests.fixtures import config_icecast


def test_import():
    """Import the reel module."""
    assert reel
    assert reel.__all__ == ['io', 'proc', 'tools']
    assert reel.io.__all__ == ['Input', 'Output', 'Through']
    assert reel.proc.__all__ == ['Daemon', 'Destination', 'Source']
    assert reel.tools.__all__ == ['resolve']


async def test_shell_commands():
    """Run a few shell commands."""

    assert 'rutabaga' in await Source(f'grep utabag {__file__}').read_text()

    found = Source('find .', xconf=['-type', 'f'])
    assert __file__ in await found.read_list(through=resolve)

    assert float((await Source('python -V').read_text())[7:10]) >= 3.5

    xconf = ['-c', "import os; print(os.environ['_ASDF'])"]
    env = {'_ASDF': 'asdf'}
    assert await Source('python', xconf=xconf, xenv=env).read_text() == 'asdf'

    # fname = 'test.flac'
    # assert await Source(f'flac -t {fname}').read_bool()
    # assert isinstance(await Source(f'flac -t {fname}').read_bool(), bool)

    # async with Source(f'flac -t {fname}') as test:
    #     assert await test.read_bool()
    #     assert test.status == 0
    #     assert 'test.flac: ok' in test.err

    # wav = await Source(f'flac -cd {fname}').read_bytes()
    # assert len(wav) == 57684956
    # flac = await Source(f'flac -c -').read_bytes(send_bytes=wav)
    # assert len(flac) == 27583002


async def test_server(config_icecast):
    """Run a server."""
    config_dir = await get_xdg_config_dir()
    config = await get_config(config_dir, 'icecast.xml', **config_icecast)
    xconf = ['-c', str(config)]
    async with Daemon('icecast', xconf=xconf) as icecast:
        await icecast.start()
        procs = await Source('ps ax').read_list()
        found = False
        for proc in procs:
            if 'icecast' in proc:
                found = True
        assert found
    procs = await Source('ps ax').read_list()
    found = False
    for proc in procs:
        if 'icecast' in proc:
            found = True
    assert not found



"""Test suite for the reel command line interface."""
from trio_click.testing import CliRunner

import reel
from reel import cli
from reel.proc import Source


async def test_no_args():
    """Run cli with no arguments."""
    runner = CliRunner()
    result = await runner.invoke(cli.main)
    assert result.exit_code == 0
    assert result.output == ''


async def test_version():
    """Show the version number."""
    runner = CliRunner()
    result = await runner.invoke(cli.main, ['--version'])
    assert result.exit_code == 0
    assert reel.__version__ in result.output


async def test_no_args_external():
    """Run the CLI externally with no arguments."""
    command = 'python -m reel.cli'
    config = {'COVERAGE_PROCESS_START': 'setup.cfg', 'PYTHONPATH': '.'}
    noargs = Source(command, config)
    output = await noargs.run()
    assert noargs.status == 0
    assert output == ''


async def test_version_external():
    """Show the version number by running the actual CLI."""
    command = 'python -m reel.cli -v'
    env = {'COVERAGE_PROCESS_START': 'setup.cfg', 'PYTHONPATH': '.'}
    version = Source(command, env)
    output = await version.read_text()
    assert version.status == 0
    assert output == reel.__version__



"""Utility functions for the reel package."""
from trio import Path

__all__ = ['resolve']


async def resolve(path):
    """Return the canonical str representation of the file path."""
    expanded = await Path(path).expanduser()
    resolved = await expanded.resolve()
    return str(resolved)

# def grep



"""Test the pre-configured commands."""
import os

from reel import cmd
from reel.cmd import ffmpeg, play


async def test_cmd_module():
    """Import the cmd module."""
    assert cmd
    assert cmd.SRC_SILENCE


async def test_play_radio():
    """Play a few seconds of music."""
    radio = 'http://ice1.somafm.com/groovesalad-256-mp3'
    music = ''.join(['https://archive.org/download/',
                     'gd1977-05-08.shure57.stevenson.29303.flac16/',
                     'gd1977-05-08d02t04.flac'])
    async with await play.speaker() as out:
        async with await ffmpeg.stream(music) as src:
            count = 0
            chunk = await src.receive_some(4096)
            while chunk:
                if os.environ.get('REEL_ALLOW_AUDIO_PLAY'):
                    await out.send_all(chunk)
                if count > 200:
                    break
                chunk = await src.receive_some(4096)
                count += 1
                assert chunk
        async with await ffmpeg.stream(radio) as src:
            count = 0
            chunk = await src.receive_some(4096)
            while chunk:
                if os.environ.get('REEL_ALLOW_AUDIO_PLAY'):
                    await out.send_all(chunk)
                if count > 200:
                    break
                chunk = await src.receive_some(4096)
                count += 1
                assert chunk



"""Various ffmpeg command line configurations."""
from reel.proc import Source


async def stream(uri):
    """Stream an audio source."""
    flags = ['-ac', '2',
             '-i', uri,
             '-f', 's16le',
             '-ar', '44.1k',
             '-acodec', 'pcm_s16le',
             '-']
    return await Source('ffmpeg', xconf=flags).stream()



"""Handle subprocesses and pipes."""
from abc import ABCMeta, abstractmethod
import os
import shlex

import trio

from reel.io import Input, Output

__all__ = ['Daemon', 'Destination', 'Source']
LIMIT = 163840


class ProcBase(metaclass=ABCMeta):
    """Base class for subprocesses."""

    def __init__(self, command, xenv=None, xconf=None):
        """Initialize the server."""
        self._command = shlex.split(command)
        if xconf:
            for flag in xconf:
                self._command.append(shlex.quote(flag))
        self._env = os.environ
        if xenv:
            for key, val in xenv.items():
                self._env[key] = val
        self._proc = None
        self._status = None

    @abstractmethod
    async def prepare_command(self):
        """Prepare a command to start the process."""
        return self._command

    async def __aenter__(self):
        """Enter the server context."""
        return self

    async def __aexit__(self, *args):
        """Exit the server context."""
        if self._proc:
            self._proc.terminate()
            await self._proc.aclose()


class Daemon(ProcBase):
    """A server process."""

    async def prepare_command(self):
        """Return the command for this process."""
        return self._command

    async def start(self):
        """Start the server."""
        command = await self.prepare_command()
        self._proc = trio.subprocess.Process(
            command,
            stdin=None,
            stdout=None,
            stderr=None,
            env=self._env
        )


class Destination:
    """A subprocess that consumes input."""

    _command = None
    _env = os.environ
    _nurse = None
    _status = None

    def __init__(self, command, xenv=None, xconf=None):
        """Get the command ready to run."""
        self._command = shlex.split(command)
        self._proc = None
        if xconf:
            for flag in xconf:
                self._command.append(flag)
        if xenv:
            for key, val in xenv.items():
                self._env[key] = val
        print(' '.join(self._command))

    async def __aenter__(self):
        """Hijack ``trio._core._run.NurseryManager``."""
        self._nurse = trio.open_nursery()
        await self._nurse.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        """Close the nursery."""
        await self._nurse.__aexit__(exc_type, exc, traceback)

    @property
    def status(self):
        """Return the command exit code."""
        return self._status

    async def receive(self):
        """Return a context-managed output stream."""
        self._proc = trio.subprocess.Process(
            self._command,
            stdin=trio.subprocess.PIPE,
            stdout=None,
            stderr=None,
            env=self._env
        )
        return Input(self._proc.stdin)


class Source():
    """A subprocess that produces output."""

    _command = None
    _env = os.environ
    _err = b''
    _limit = LIMIT
    _nurse = None
    _output = b''
    _proc = None
    _status = None
    _stream = False

    def __init__(self, command, xenv=None, xconf=None):
        """Get the command ready to run."""
        self._command = shlex.split(command)
        if xconf:
            for flag in xconf:
                self._command.append(flag)
        if xenv:
            for key, val in xenv.items():
                self._env[key] = val
        print(' '.join(self._command))

    async def __aenter__(self):
        """Hijack ``trio._core._run.NurseryManager``."""
        self._nurse = trio.open_nursery()
        await self._nurse.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        """Close the nursery."""
        await self._nurse.__aexit__(exc_type, exc, traceback)

    @property
    def cmd(self):
        """Return the original command."""
        return ' '.join(self._command)

    @property
    def err(self):
        """Decode stderr."""
        result = []
        decoded = self._err.decode('utf-8').strip()
        for byte in decoded:
            if byte == '':
                result.pop()
            else:
                result.append(byte)
        return ''.join(result)

    @property
    def status(self):
        """Return the command exit code."""
        return self._status

    async def _stream_stderr(self, limit):
        """..."""
        async with self._proc.stderr as err:  # type: trio.abc.ReceiveStream
            while True:
                chunk = await err.receive_some(limit)
                if not chunk:
                    break
                self._err += chunk
                if limit:
                    if len(self._err) >= limit:
                        break

    async def _stream_stdout(self, limit):
        """..."""
        async with self._proc.stdout as out:
            while True:
                if limit:
                    chunk_size = limit
                else:
                    chunk_size = 16384
                chunk = await out.receive_some(chunk_size)
                if not chunk:
                    break
                self._output += chunk
                if limit:
                    if len(self._output) >= limit:
                        break

    async def _stream_stdin(self, message):
        """..."""
        async with self._proc.stdin as stdin:
            try:
                await stdin.send_all(message)
            except trio.BrokenResourceError:
                pass

    async def run(self, message=None):
        """Run the command."""
        if message:
            message = message.encode('utf-8')
            _stdin = trio.subprocess.PIPE
        else:
            _stdin = None
        self._proc = trio.subprocess.Process(
            self._command,
            stdin=_stdin,
            stdout=None,
            stderr=None,
            env=self._env
        )
        if message:
            async with trio.open_nursery() as nursery:
                nursery.start_soon(self._stream_stdin, message)
        await self._proc.wait()
        self._status = self._proc.returncode
        return self._output.decode('utf-8').strip()

    async def read_bool(self, stdin=None):
        """Run the command and return the exit status."""
        if stdin:
            stdin = stdin.encode('utf-8')
            _stdin = trio.subprocess.PIPE
        else:
            _stdin = None
        self._proc = trio.subprocess.Process(
            self._command,
            stdin=_stdin,
            stdout=trio.subprocess.PIPE,
            stderr=trio.subprocess.PIPE,
            env=self._env
        )
        async with trio.open_nursery() as nursery:
            if _stdin == trio.subprocess.PIPE:
                nursery.start_soon(self._stream_stdin, stdin)
            nursery.start_soon(self._stream_stderr, self._limit)
            nursery.start_soon(self._stream_stdout, self._limit)
            await self._proc.wait()
            self._status = self._proc.returncode
        return self.status == 0

    async def read_text(self, message=None):
        """Run the command and return the output."""
        if message:
            message = message.encode('utf-8')
            _stdin = trio.subprocess.PIPE
        else:
            _stdin = None
        self._proc = trio.subprocess.Process(
            self._command,
            stdin=_stdin,
            stdout=trio.subprocess.PIPE,
            stderr=trio.subprocess.PIPE,
            env=self._env
        )
        async with trio.open_nursery() as nursery:
            if _stdin == trio.subprocess.PIPE:
                nursery.start_soon(self._stream_stdin, message)
            nursery.start_soon(self._stream_stderr, self._limit)
            nursery.start_soon(self._stream_stdout, self._limit)
            await self._proc.wait()
            self._status = self._proc.returncode
        return self._output.decode('utf-8').strip()

    async def read_bytes(self, send_bytes=None):
        """Run the command and return the output."""
        if send_bytes:
            _stdin = trio.subprocess.PIPE
        else:
            _stdin = None
        self._proc = trio.subprocess.Process(
            self._command,
            stdin=_stdin,
            stdout=trio.subprocess.PIPE,
            stderr=trio.subprocess.PIPE,
            env=self._env
        )
        async with trio.open_nursery() as nursery:
            if _stdin == trio.subprocess.PIPE:
                nursery.start_soon(self._stream_stdin, send_bytes)
            nursery.start_soon(self._stream_stderr, self._limit)
            nursery.start_soon(self._stream_stdout, None)
            await self._proc.wait()
            self._status = self._proc.returncode
        return self._output

    async def read_list(self, message=None, through=None):
        """Return the output as a list."""
        output = await self.read_text(message)
        output_list = output.split('\n')
        if through:
            return [await through(_) for _ in output_list]
        return output_list

    async def stream(self, message=None):
        """Return a context-managed output stream."""
        if message:
            message = message.encode('utf-8')
            _stdin = trio.subprocess.PIPE
        else:
            _stdin = None
        self._proc = trio.subprocess.Process(
            self._command,
            stdin=_stdin,
            stdout=trio.subprocess.PIPE,
            stderr=None,
            env=self._env
        )
        if message:
            async with trio.open_nursery() as nursery:
                nursery.start_soon(self._stream_stdin, message)
        return Output(self._proc.stdout)
